/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_ttf.h>
#include <png.h>
#include <stdarg.h>

#include "gp2x_psp.h"
#include "psp_sdl.h"
#include "psp_global.h"
#include "psp_kbd.h"
#include "psp_chess.h"
#include "psp_menu.h"
#include "psp_play.h"

  static int psp_chess_board_dirty = 1;
  static int psp_chess_move_mode   = 1;
  static int psp_chess_status_mode = 0;

  static char xboard_buffer[128];


void
psp_chess_set_move_mode(int mode)
{
  psp_chess_move_mode = mode;
}

static inline void
psp_chess_get_pix_xy(int x, int y, int *px_pix, int *py_pix)
{
  int square_width_pix = 240 / (GCHESS_BOARD_WIDTH + 1);
  int board_width_pix  = square_width_pix * (GCHESS_BOARD_WIDTH - 1);
  int x_orig = (360 - board_width_pix) / 2;
  int y_orig = (240 - board_width_pix) / 2;
  
  *px_pix = x_orig + (x * square_width_pix);
  *py_pix = y_orig + ((GCHESS_BOARD_WIDTH - 1 - y) * square_width_pix);
}

int
psp_chess_is_computer_move()
{
  if ((! psp_chess_move_mode) || psp_chess_is_game_over()) return 0;

  if (GCHESS.chess_game_mode == GCHESS_HUMAN_VS_COMPUTER) {
    return (GCHESS.chess_player_color != board.side);
  } else 
  if (GCHESS.chess_game_mode == GCHESS_COMPUTER_VS_COMPUTER) {
    return 1;
  }
  return 0;
}

int
psp_chess_get_pieced_xy(int x, int y, int* is_black, int* piece)
{
  int r;
  int c;
  int sq;

  if (GCHESS.chess_side_color == white) {
    r = 56 - y * 8;
    c = x;
    sq = r + c;
  } else {
    r = y * 8;
    c = 7 - x;
    sq = r + c;
  }

  /* white = 0   black = 1 */
  int a_color = 0;
  for (a_color = 0; a_color < 2; a_color++) {
    *is_black = (a_color == black);
    if (board.b[a_color][pawn] & BitPosArray[sq]) {
      *piece = CHESS_PAWN;
      return 1;
    } else 
    if (board.b[a_color][knight] & BitPosArray[sq]) {
      *piece = CHESS_KNIGHT;
      return 1;
    } else 
    if (board.b[a_color][bishop] & BitPosArray[sq]) {
      *piece = CHESS_BISHOP;
      return 1;
    } else 
    if (board.b[a_color][rook] & BitPosArray[sq]) {
      *piece = CHESS_ROOK;
      return 1;
    } else 
    if (board.b[a_color][queen] & BitPosArray[sq]) {
      *piece = CHESS_QUEEN;
      return 1;
    } else 
    if (board.b[a_color][king] & BitPosArray[sq]) {
      *piece = CHESS_KING;
      return 1;
    }
  }
  return 0;
}

void
psp_chess_cursor_reset()
{
  GCHESS.move_x[0] = 0;
  GCHESS.move_y[0] = 0;
  GCHESS.move_y[0] = 7;
  GCHESS.move_x[1] = 0;
  GCHESS.move_y[1] = 0;
  GCHESS.move_step = 0;
}

void
psp_chess_change_color(int new_color)
{
  GCHESS.chess_player_color = new_color;
  if (computer == new_color) {
    computer = ! new_color;

    CLEAR (flags, MANUAL);
    CLEAR (flags, TIMEOUT);
    CLEAR (flags, ENDED);

    if (board.side == computer) {
      cmd_go();
    }
  }
  psp_chess_cursor_reset();
}

void 
psp_chess_change_search_depth(int search_depth) 
{
  GCHESS.chess_search_depth = search_depth;
  SearchDepth = search_depth;
}

# if 0 //LUDO: TO_BE_DONE 
void 
psp_chess_change_limit(int time_limit, int move_limit) 
{
  GCHESS.chess_time_limit = time_limit;
  GCHESS.chess_move_limit = move_limit;

  TCTime = time_limit;
  TCMove = move_limit;

  if (TCTime || TCMove) SET (flags, TIMECTL);
  else                  CLEAR (flags, TIMECTL);

  TimeLimit[white] = TimeLimit[black] = TCTime * 60;
  if (TCMove) {
    MoveLimit[white] = MoveLimit[black] = TCMove - (GameCnt+1)/2;
  } else {
    MoveLimit[white] = MoveLimit[black] = 0;
  }
}
# endif

void 
psp_chess_change_book(int book) 
{
  GCHESS.chess_book = book;
  bookmode = book;
}

int
psp_chess_get_last_move(int* p_last_x, int* p_last_y)
{
  if (GameCnt <= 0) return 0;

  int sq = TOSQ(Game[GameCnt].move);

  if (GCHESS.chess_side_color == white) {
    sq = 63 - sq;
  }
  *p_last_x = 7 - (sq % 8);
  *p_last_y = sq / 8;

  return 1;
}

void
psp_chess_display_board()
{
  int  x_orig;
  int  y_orig;
  int  x_dst;
  int  y_dst;
  int  x;
  int  y;

  int square_width_pix = 30;
  int board_width_pix  = square_width_pix * 8;

  x_orig = (320 - board_width_pix) / 2;
  y_orig = (240 - board_width_pix) / 2;

  psp_sdl_blit_board();

  psp_sdl_draw_rectangle(x_orig, y_orig, board_width_pix, board_width_pix, 0, 0);

  int w_color = GCHESS.white_color;
  int b_color = GCHESS.black_color;
  int sel_color = GCHESS.select_color;
  int last_color = GCHESS.last_color;
  int valid_color = GCHESS.valid_color;
  int color = 0;

  y_dst = y_orig;
  for (y = 0; y < GCHESS_BOARD_WIDTH; y++) {
    x_dst = x_orig;
    for (x = 0; x < GCHESS_BOARD_WIDTH; x++) {
      if ((x & 1) == (y & 1)) color = w_color;
      else                    color = b_color;
      if (GCHESS.move_step) {
        if ((x == GCHESS.move_x[0]) && (y == GCHESS.move_y[0])) color = valid_color;
      } 
      psp_sdl_fill_rectangle(x_dst, y_dst, square_width_pix, square_width_pix, color, 0);
      psp_sdl_draw_rectangle(x_dst, y_dst, square_width_pix, square_width_pix, 0, 0);

      int is_black = 0;
      int a_piece  = 0;
      if (psp_chess_get_pieced_xy(x, y, &is_black, &a_piece)) {
        psp_sdl_blit_piece_xy(x_dst, y_dst, is_black, a_piece);
      }
      x_dst += square_width_pix;
    }
    y_dst += square_width_pix;
  }

  if (GCHESS.chess_show_coord) {
    y_dst = y_orig + (square_width_pix - 8) / 2;
    x_dst = x_orig - 10;
    for (y = 0; y < GCHESS_BOARD_WIDTH; y++) {
      if (GCHESS.chess_side_color == white) {
        psp_sdl_back3_put_char(x_dst, y_dst, sel_color, '8' - y );
      } else {
        psp_sdl_back3_put_char(x_dst, y_dst, sel_color, y + '1' );
      }
      y_dst += square_width_pix;
    }
    //y_dst = y_orig + (square_width_pix * GCHESS_BOARD_WIDTH) + 4;
    y_dst = y_orig + (square_width_pix * GCHESS_BOARD_WIDTH) - square_width_pix;
    x_dst = x_orig + square_width_pix - 7;
    for (x = 0; x < GCHESS_BOARD_WIDTH; x++) {
      if (GCHESS.chess_side_color == white) {
        psp_sdl_put_char(x_dst, y_dst, valid_color, 0, x + 'a', 1, 0 );
      } else {
        psp_sdl_put_char(x_dst, y_dst, valid_color, 0, 'h' - x, 1, 0 );
      }
      x_dst += square_width_pix;
    }
  }

  if (GCHESS.chess_last_move) {

    int last_x = 0;
    int last_y = 0;

    if (psp_chess_get_last_move(&last_x, &last_y)) {
      x_dst = x_orig + last_x * square_width_pix;
      y_dst = y_orig + last_y * square_width_pix;

      psp_sdl_draw_rectangle(x_dst, y_dst, square_width_pix, square_width_pix, last_color, 0);
      psp_sdl_draw_rectangle(x_dst+1, y_dst+1, square_width_pix-2, square_width_pix-2, last_color, 0);
    }
  }

  if (! psp_chess_is_computer_move()) {

    int current_x = GCHESS.move_x[GCHESS.move_step];
    int current_y = GCHESS.move_y[GCHESS.move_step];
    x_dst = x_orig + current_x * square_width_pix;
    y_dst = y_orig + current_y * square_width_pix;
    psp_sdl_draw_rectangle(x_dst, y_dst, square_width_pix, square_width_pix, sel_color, 0);
    psp_sdl_draw_rectangle(x_dst+1, y_dst+1, square_width_pix-2, square_width_pix-2, sel_color, 0);

  }
}

void
psp_chess_display_status()
{
  char buffer[128];

  int b_b_color = GCHESS.black_color;
  int f_w_color = GP2X_MENU_WHITE_COLOR;
  int f_b_color = psp_sdl_rgb(0, 0, 0);
  int menu_color = GP2X_MENU_WHITE_COLOR;
  int select_color = GCHESS.select_color;

  int b_x    =   0;
  int width  = 120;
  int height = 200;

  int y_orig = (240 - height) / 2;

  psp_sdl_draw_rectangle( b_x    ,  y_orig    , width    , height    , 0, 0);
  psp_sdl_fill_rectangle( b_x + 1,  y_orig + 1, width - 2, height - 2, b_b_color, 0);

  b_x    += 5;
  y_orig += 10;

  if (board.side == black) {
    psp_sdl_print(b_x, y_orig, "Black to play", f_b_color);
  } else {
    psp_sdl_print(b_x, y_orig, "White to play", f_w_color);
  }
  y_orig += 10;
  y_orig += 10;

  if (psp_chess_is_computer_move()) {
    psp_sdl_print(b_x, y_orig, " GP2X is thinking", select_color);
  } else 
  if (! psp_chess_is_game_over()) {
    psp_sdl_print(b_x, y_orig, " Select your move", select_color);
  } else {
    psp_sdl_print(b_x, y_orig, " Game over ...", select_color);
  }
  y_orig += 10;
  y_orig += 10;

  psp_sdl_print(b_x, y_orig, "Last comment :", menu_color);
  y_orig += 10;
  y_orig += 10;
  snprintf(buffer, 20, " %s", xboard_buffer);
  string_fill_with_space(buffer, 20);
  psp_sdl_print(b_x, y_orig, buffer, menu_color);
  y_orig += 10;
  y_orig += 10;

  psp_sdl_print(b_x, y_orig, "Last moves :", menu_color);
  y_orig += 10;
  y_orig += 10;

  int index;
  for (index = 0; index < 8; index++) {
    strcpy(buffer, "    .");
    int abs_index = GameCnt - index;
    if (abs_index >= 0) {
      char* move_str = ExtraAlgbrMove(Game[abs_index].move);
      sprintf(buffer, " %3d. %s", abs_index + 1, move_str);
    }
    string_fill_with_space(buffer, 18);
    if (abs_index & 1) {
      psp_sdl_print(b_x, y_orig, buffer, f_b_color);
    } else {
      psp_sdl_print(b_x, y_orig, buffer, f_w_color);
    }
    y_orig += 10;
  }
# if 0 //LUDO: TO_BE_DONE !
  y_orig += 10;
  sprintf(buffer, "White : %d %d", (int)TimeLimit[white], MoveLimit[white]);
  psp_sdl_print(b_x, y_orig, buffer, menu_color);
  y_orig += 10;
  sprintf(buffer, "Black : %d %d", (int)TimeLimit[black], MoveLimit[black]);
  psp_sdl_print(b_x, y_orig, buffer, menu_color);
# endif
}

void
psp_chess_refresh_screen()
{
  psp_chess_display_board();
  if (psp_chess_status_mode) {
    psp_chess_display_status();
  }
}

void
psp_chess_chess_up()
{
  GCHESS.move_y[GCHESS.move_step]--; 
  if (GCHESS.move_y[GCHESS.move_step] < 0) GCHESS.move_y[GCHESS.move_step] = GCHESS_BOARD_WIDTH - 1;
  psp_chess_board_dirty = 1;
}

void
psp_chess_chess_down()
{
  GCHESS.move_y[GCHESS.move_step]++; 
  if (GCHESS.move_y[GCHESS.move_step] >= GCHESS_BOARD_WIDTH) GCHESS.move_y[GCHESS.move_step] = 0;
  psp_chess_board_dirty = 1;
}

void
psp_chess_chess_left()
{
  GCHESS.move_x[GCHESS.move_step]--; 
  if (GCHESS.move_x[GCHESS.move_step] < 0) GCHESS.move_x[GCHESS.move_step] = GCHESS_BOARD_WIDTH - 1;
  psp_chess_board_dirty = 1;
}

void
psp_chess_chess_right()
{
  GCHESS.move_x[GCHESS.move_step]++; 
  if (GCHESS.move_x[GCHESS.move_step] >= GCHESS_BOARD_WIDTH) GCHESS.move_x[GCHESS.move_step] = 0;
  psp_chess_board_dirty = 1;
}

# define ANALOG_THRESHOLD 30

void 
psp_chess_get_analog_direction(int Analog_x, int Analog_y, int *x, int *y)
{
  int DeltaX = 255;
  int DeltaY = 255;
  int DirX   = 0;
  int DirY   = 0;

  *x = 0;
  *y = 0;

  if (Analog_x <=        ANALOG_THRESHOLD)  { DeltaX = Analog_x; DirX = -1; }
  else 
  if (Analog_x >= (255 - ANALOG_THRESHOLD)) { DeltaX = 255 - Analog_x; DirX = 1; }

  if (Analog_y <=        ANALOG_THRESHOLD)  { DeltaY = Analog_y; DirY = -1; }
  else 
  if (Analog_y >= (255 - ANALOG_THRESHOLD)) { DeltaY = 255 - Analog_y; DirY = 1; }

  *x = DirX;
  *y = DirY;
}

void
psp_chess_new_game()
{
  psp_global_reset();
}

int
psp_chess_is_game_over()
{
  return flags & ENDED;
}

int
xboard_printf(const char* format, ...)
{
  va_list ap;
  va_start(ap, format);
  int i = vsnprintf(xboard_buffer, sizeof(xboard_buffer), format, ap);
  va_end(ap);
  return i;
}

void
psp_chess_forward_move()
{
# if 0 //LUDO: 
  if (sgftreeForward(&GCHESS.sgftree)) {
    GCHESS.gameinfo.to_move = gnuchess_play_sgfnode(GCHESS.sgftree.lastnode, GCHESS.gameinfo.to_move);
    psp_chess_board_dirty = 1;
  }
# endif
}

void
psp_chess_undo_move()
{
  if (GCHESS.chess_game_mode == GCHESS_HUMAN_VS_COMPUTER) {
    if (GameCnt >= 1) {
      cmd_undo();
      cmd_undo();
    } else {
      xboard_printf("No moves to undo !");
    }

  } else
  if (GCHESS.chess_game_mode == GCHESS_HUMAN_VS_HUMAN) {
    cmd_undo();
  }
  psp_chess_board_dirty = 1;
}

int
psp_chess_is_promotion()
{
  int is_black = 0;
  int a_piece = 0;
  int found = psp_chess_get_pieced_xy(GCHESS.move_x[0], GCHESS.move_y[0], &is_black, &a_piece);
  if (found) {
    if (a_piece == CHESS_PAWN) {
      if ((GCHESS.move_y[1] == 0) || (GCHESS.move_y[1] == 7)) {
        return 1;
      }
    }
  }
  return 0;
}

void
psp_chess_move()
{
  if (psp_chess_is_game_over()) return;

  int is_black = 0;
  int a_piece  = 0;
  int back_to_step_0 = 0;

  int found = psp_chess_get_pieced_xy(GCHESS.move_x[GCHESS.move_step], 
                                      GCHESS.move_y[GCHESS.move_step], &is_black, &a_piece);
  
  if (GCHESS.chess_game_mode != GCHESS_COMPUTER_VS_COMPUTER) {
    if (GCHESS.move_step == 0) {
      if (found) {
        /* Check if pieced at (x,y) is of player color */
        if (is_black == board.side) {
          GCHESS.move_step++;
          GCHESS.move_x[1] = GCHESS.move_x[0];
          GCHESS.move_y[1] = GCHESS.move_y[0];
        }
      }
    } else {
      int do_move = 0;

      if (! found) {
        do_move = 1;
      } else {
        /* Check if move_0 and move_1 are equal */
        if ((GCHESS.move_x[0] == GCHESS.move_x[1]) &&
            (GCHESS.move_y[0] == GCHESS.move_y[1])) {
          back_to_step_0 = 1;
        } else
        /* Check if pieced at (x,y) is of player color */
        if (is_black == board.side) {
          GCHESS.move_step = 1;
          GCHESS.move_x[0] = GCHESS.move_x[1];
          GCHESS.move_y[0] = GCHESS.move_y[1];
        } else {
          do_move = 1;
        }
      }

      if (do_move) {
        if (GCHESS.chess_side_color == white) {
          sprintf(inputstr, "%c%d%c%d", GCHESS.move_x[0] + 'a', 8 - GCHESS.move_y[0], 
                                        GCHESS.move_x[1] + 'a', 8 - GCHESS.move_y[1]);
        } else {
          sprintf(inputstr, "%c%d%c%d", 'h' - GCHESS.move_x[0], GCHESS.move_y[0] + 1, 
                                        'h' - GCHESS.move_x[1], GCHESS.move_y[1] + 1);
        }

        if (psp_chess_is_promotion()) {
          strcat(inputstr, "q");
        }
        xboard_buffer[0] = 0;

        if (! parse_input()) {
          /* nothing to do */
        } else {
          /* illegal move ... */
        }

        psp_chess_refresh_screen();
        psp_sdl_flip();

        back_to_step_0 = 1;
      }
    }
    if (back_to_step_0) {
      /* Back to step 0 */
      GCHESS.move_step = 0;
      GCHESS.move_x[0] = GCHESS.move_x[1];
      GCHESS.move_y[0] = GCHESS.move_y[1];
    }
    psp_chess_board_dirty = 1;
  }
}

void
psp_chess_resign_game()
{
  SET(flags, TIMEOUT | ENDED);
}

int
psp_chess_is_key_pressed()
{
  gp2xCtrlData c;
  gp2xCtrlPeekBufferPositive(&c, 1);
  c.Buttons &= GP2X_ALL_BUTTON_MASK;
  return c.Buttons;
}

void
psp_chess_main_loop()
{
  gp2xCtrlData c;
  long        new_pad = 0;
  long        old_pad;
  int         last_time;

  int new_Lx = 0;
  int new_Ly = 0;

  old_pad   = 0;
  last_time = 0;

  psp_main_menu();

  while (1) {

    do {

      if (GCHESS.psp_screenshot_mode) {
        GCHESS.psp_screenshot_mode = 0;
        psp_chess_refresh_screen();
        psp_sdl_flip();
        psp_chess_refresh_screen();
        psp_sdl_flip();
        psp_sdl_save_screenshot();
      }

      if (psp_chess_board_dirty) {
        psp_chess_board_dirty = 0;
        psp_chess_refresh_screen();
        psp_sdl_flip();
      }

      if (! psp_chess_is_game_over()) {

        if (psp_chess_is_computer_move()) {

          xboard_buffer[0] = 0;

          psp_kbd_wait_no_button();
          psp_play_computer_move();

          psp_chess_refresh_screen();
          psp_sdl_flip();

          if (GCHESS.chess_game_mode == GCHESS_COMPUTER_VS_COMPUTER) {
            cmd_go();
          }
        }
      }

    } while (psp_chess_board_dirty);

    if (new_pad & (GP2X_CTRL_SQUARE|GP2X_CTRL_CROSS|GP2X_CTRL_CIRCLE|GP2X_CTRL_TRIANGLE)) {
      psp_kbd_wait_no_button();
    }

    while (1)
    {
      gp2xCtrlPeekBufferPositive(&c, 1);

      if (c.Buttons)  break;
      if (GCHESS.chess_game_mode == GCHESS_COMPUTER_VS_COMPUTER) {
        break;
      }
    }

    new_pad = c.Buttons;

    if (new_pad) {
      if ((old_pad != new_pad) || ((c.TimeStamp - last_time) > GP2X_CHESS_MIN_TIME)) {
        last_time = c.TimeStamp;
        old_pad = new_pad;
      } else continue;

    } else {
      if ((c.TimeStamp - last_time) > GP2X_CHESS_MIN_ANALOG_TIME) {
        last_time = c.TimeStamp;
      } else continue;
    }

    if ((c.Buttons & (GP2X_CTRL_LTRIGGER|GP2X_CTRL_RTRIGGER|GP2X_CTRL_START)) ==
        (GP2X_CTRL_LTRIGGER|GP2X_CTRL_RTRIGGER|GP2X_CTRL_START)) {
      /* Exit ! */
      psp_sdl_exit(0);
    } else
    if((new_pad & GP2X_CTRL_SELECT) == GP2X_CTRL_SELECT) {
      psp_main_menu();
      psp_chess_board_dirty = 1;
    } else {

      if (new_pad & GP2X_CTRL_CROSS) {
        psp_chess_move_mode = 1;
        psp_chess_move();
      } else
      if (new_pad & GP2X_CTRL_CIRCLE) {
        psp_chess_move_mode = 1;
        psp_chess_move();
      } else
      if ((new_pad & GP2X_CTRL_LTRIGGER) == GP2X_CTRL_LTRIGGER) {
        psp_chess_status_mode = ! psp_chess_status_mode;
        psp_chess_board_dirty = 1;
      } else
      if (new_pad & GP2X_CTRL_SQUARE) {
        psp_chess_move_mode = 0;
        psp_chess_undo_move();
      } else
      if ((new_pad & GP2X_CTRL_RTRIGGER) == GP2X_CTRL_RTRIGGER) {
        psp_chess_status_mode = ! psp_chess_status_mode;
        psp_chess_board_dirty = 1;
      } else
      if (new_pad & GP2X_CTRL_TRIANGLE) {
        psp_chess_move_mode = 0;
        psp_chess_forward_move();
      }
      if ((new_pad & GP2X_CTRL_UP) || (new_Ly < 0)) {
        psp_chess_chess_up();
      }
      if ((new_pad & GP2X_CTRL_DOWN) || (new_Ly > 0)) {
        psp_chess_chess_down();
      }
      if ((new_pad & GP2X_CTRL_LEFT) || (new_Lx < 0)) {
        psp_chess_chess_left();
      }
      if ((new_pad & GP2X_CTRL_RIGHT) || (new_Lx > 0)) {
        psp_chess_chess_right();
      }
    }
  }
}

int
psp_chess_load_pgn(char *filename)
{
  int error = PGNReadFromFile(filename);
  if (! error) {
    char* scan = strrchr(filename, '/');
    if (scan) {
      strcpy(GCHESS.game_filename, scan+1);
    } else {
      strcpy(GCHESS.game_filename, filename);
    }
  }
  return error;
}

int
psp_chess_save_pgn(char *pgn_filename)
{
  return PGNSaveToFile(pgn_filename, NULL, 0);
}

int
psp_chess_append_pgn(char *pgn_filename)
{
  return PGNSaveToFile(pgn_filename, NULL, 1);
}

int 
SDL_main()
{
  memset(&GCHESS, 0, sizeof(GCHESS));
  getcwd(GCHESS.psp_homedir, sizeof(GCHESS.psp_homedir));
  psp_sdl_init();
  gnuchess_init();
  psp_global_init();
  psp_chess_main_loop();
  gnuchess_end();

  psp_sdl_exit(0);

  return 0;
}

